use itertools::Itertools;
use json::JsonValue;
use log::{debug, error};
use std::fmt::Debug;

pub trait Transform: Debug {
    fn applies_to(&self, topic: &str) -> bool;
    fn transform(
        &self,
        topic: &str,
        data: JsonValue,
        labels: &mut Vec<(String, String)>,
    ) -> (String, JsonValue);
}

#[derive(Debug)]
pub struct ZwaveNormalize {
    topic_prefix: String,
}

impl ZwaveNormalize {
    pub fn new(topic_prefix: &str) -> Self {
        Self {
            topic_prefix: topic_prefix.to_owned(),
        }
    }
}

impl Transform for ZwaveNormalize {
    fn applies_to(&self, topic: &str) -> bool {
        topic.starts_with(&self.topic_prefix)
    }
    fn transform(
        &self,
        topic: &str,
        mut data: JsonValue,
        labels: &mut Vec<(String, String)>,
    ) -> (String, JsonValue) {
        if topic.contains("nodeinfo") || !topic.contains("endpoint_") {
            debug!("zwave topic isn't relevant");
            return (topic.to_owned(), JsonValue::Null);
        }
        if !data.is_object() || !data.has_key("value") {
            debug!("zwave payload doesn't include value");
            return (topic.to_owned(), JsonValue::Null);
        }
        debug!("Normalizing zwave2mqtt payload");
        let parts = topic.split('/').collect_vec();
        if let Some((properties_index, _)) = parts
            .iter()
            .enumerate()
            .find(|(_, v)| v.starts_with("endpoint_"))
        {
            let parts = topic.split('/').collect_vec();
            let topic = parts[..properties_index]
                .iter()
                .map(|s| s.to_lowercase())
                .join("/");

            if let Ok(endpoint) = parts[properties_index]
                .split_at("endpoint_".len())
                .1
                .parse::<u64>()
            {
                labels.push(("endpoint".to_owned(), endpoint.to_string()));
            } else {
                labels.push(("endpoint".to_owned(), parts[properties_index].to_owned()))
            }

            let properties = parts[properties_index + 1..]
                .iter()
                .map(|s| s.to_lowercase())
                .join("_");
            let mut value = JsonValue::new_object();
            value.insert(&properties, data.remove("value")).unwrap();

            debug!("Normalized topic: {topic} value: {value}");

            (topic, value)
        } else {
            error!(target: "ZwaveNormalize", "Unable to find propreties index topic: {topic:?} value: {data:?}");
            (topic.to_owned(), JsonValue::Null)
        }
    }
}

#[derive(Debug)]
pub struct EsphomeNormalize {
    topic_prefixes: Vec<String>,
}
impl EsphomeNormalize {
    pub fn new(topic_prefixes: &[String]) -> Self {
        Self {
            topic_prefixes: topic_prefixes.to_vec(),
        }
    }
}
// Not sure why you'd want this, it de-normalizes the values since it moves them to separate topics per esphome node, but okay...
impl Transform for EsphomeNormalize {
    fn applies_to(&self, topic: &str) -> bool {
        self.topic_prefixes.iter().any(|pfx| topic.starts_with(pfx))
    }
    fn transform(
        &self,
        topic: &str,
        data: JsonValue,
        labels: &mut Vec<(String, String)>,
    ) -> (String, JsonValue) {
        // <TOPIC_PREFIX>/<COMPONENT_TYPE>/<COMPONENT_NAME>/state
        let parts = topic.split('/').collect_vec();

        labels.push(("esphome_type".to_owned(), parts[1].to_owned()));
        labels.push(("esphome_name".to_owned(), parts[2].to_owned()));

        let topic = format!("{}/{}", parts[0].to_lowercase(), parts[1].to_lowercase());
        let mut value = JsonValue::new_object();
        value.insert(parts[parts.len() - 2], data).unwrap();
        (topic, value)
    }
}

#[derive(Debug)]
pub struct HubitatNormalize {
    topic_prefixes: Vec<String>,
}

impl HubitatNormalize {
    pub fn new(topic_prefixes: &[String]) -> Self {
        Self {
            topic_prefixes: topic_prefixes.to_vec(),
        }
    }
}

impl Transform for HubitatNormalize {
    fn applies_to(&self, topic: &str) -> bool {
        self.topic_prefixes.iter().any(|pfx| topic.starts_with(pfx))
    }
    fn transform(
        &self,
        topic: &str,
        data: JsonValue,
        labels: &mut Vec<(String, String)>,
    ) -> (String, JsonValue) {
        // hubitat/hub1/some room/temperature/value
        let parts = topic.split('/').collect_vec();
        if parts.len() < 3 {
            return (topic.to_owned(), data);
        }

        let topic = parts[0..3].iter().map(|s| s.to_lowercase()).join("_");
        labels.push(("hubitat_hub".to_owned(), parts[1].to_owned()));
        labels.push(("hubitat_room".to_owned(), parts[2].to_owned()));

        let mut value = JsonValue::new_object();
        value.insert(parts[parts.len() - 2], data).unwrap();
        (topic, value)
    }
}

const ZIGBEE2MQTT_AVAILABILITY_SUFFIX: &str = "availability";

#[derive(Debug, Default)]
pub struct Zigbee2MqttAvailability {}

impl Transform for Zigbee2MqttAvailability {
    fn applies_to(&self, topic: &str) -> bool {
        topic.ends_with(ZIGBEE2MQTT_AVAILABILITY_SUFFIX)
    }
    fn transform(
        &self,
        topic: &str,
        mut data: JsonValue,
        _labels: &mut Vec<(String, String)>,
    ) -> (String, JsonValue) {
        if data.has_key("state") {
            let topic = topic
                .split_at(topic.len() - ZIGBEE2MQTT_AVAILABILITY_SUFFIX.len() - 1)
                .0;
            let mut value = JsonValue::new_object();
            value
                .insert("zigbee_availability", data.remove("state"))
                .unwrap();
            (topic.to_owned(), value)
        } else {
            (topic.to_owned(), data)
        }
    }
}
